#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['bit_cast__general', 'bit_cast__float_to_uint32', 'bit_cast__uint32_to_float']


from struct import pack, unpack
from typing import Any


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def bit_cast__general(data: Any, bit_cast_from: str, bit_cast_to: str) -> Any:
    """See https://docs.python.org/3/library/struct.html

    Args:
        data (Any): _description_
        bit_cast_from (str): see https://docs.python.org/3/library/struct.html#format-characters
        bit_cast_to (str): see https://docs.python.org/3/library/struct.html#format-characters

    Returns:
        Any: _description_
    """
    return unpack(bit_cast_to, pack(bit_cast_from, data))


def bit_cast__float_to_uint32(x: float) -> int:
    return unpack('I', pack('f', x))


def bit_cast__uint32_to_float(x: int) -> float:
    return unpack('f', pack('I', x))


def bit_cast__double_to_uint64(x: float) -> int:
    return unpack('Q', pack('d', x))


def bit_cast__uint64_to_double(x: int) -> float:
    return unpack('d', pack('Q', x))
