# compile_time_py_definitions.nim

const NIMD_UNAME_SYSNAME* = """Linux"""
const NIMD_UNAME_MACHINE* = """x86_64"""
const NIMD_CPU_ARCH* = """X86_64"""
const NIMD_CPU_ARCH_RAW_STRING* = """x86_64"""
const NIMD_CPU_BITS* = 64
const NIMD_IS_X86_BOOL* = true
const NIMD_IS_ARM_BOOL* = false
const NIMD_IS_X86* = 1
const NIMD_IS_ARM* = 0
const NIMD_COMPILER_TYPE* = """gcc"""
const NIMD_COMPILER_NAME* = """gcc"""
const NIMD_COMPILER_STRING* = """gcc"""
const NIMD_PYTHON_IMPLEMENTATION* = """CPython"""
const NIMD_PYTHON_VERSION_STR* = """3.12.3"""
const NIMD_PYTHON_VERSION_MAJOR_STR* = """3"""
const NIMD_PYTHON_VERSION_MINOR_STR* = """12"""
const NIMD_PYTHON_VERSION_MICRO_STR* = """3"""
const NIMD_PYTHON_VERSION_MAJOR* = 3
const NIMD_PYTHON_VERSION_MINOR* = 12
const NIMD_PYTHON_VERSION_MICRO* = 3
const NIMD_IS_RUNNING_IN_PYCHARM_BOOL* = false
const NIMD_IS_RUNNING_IN_PYCHARM* = 0
const NIMD_RAW_OS_PLATFORM* = """linux"""
const NIMD_OS_API_TYPE* = """posix"""
const NIMD_OS_TYPE* = """Linux"""
const NIMD_KIVY_PLATFORM* = """unknown"""
const NIMD_KIVY_TARGET_PLATFORM* = """unknown"""
const NIMD_IS_RUNNING_IN_EMSCRIPTEN_BOOL* = false
const NIMD_IS_RUNNING_IN_WASI_BOOL* = false
const NIMD_IS_RUNNING_IN_PYODIDE_BOOL* = false
const NIMD_IS_BUILDING_FOR_PYODIDE_BOOL* = false
const NIMD_IS_INSIDE_OR_FOR_WEB_BROWSER_BOOL* = false
const NIMD_IS_RUNNING_IN_EMSCRIPTEN* = 0
const NIMD_IS_RUNNING_IN_WASI* = 0
const NIMD_IS_RUNNING_IN_PYODIDE* = 0
const NIMD_IS_BUILDING_FOR_PYODIDE* = 0
const NIMD_IS_INSIDE_OR_FOR_WEB_BROWSER* = 0
const NIMD_CENGAL_VERSION_STR* = """4.4.0"""
const NIMD_CENGAL_VERSION_MAJOR_STR* = """4"""
const NIMD_CENGAL_VERSION_MINOR_STR* = """4"""
const NIMD_CENGAL_VERSION_MICRO_STR* = """0"""
const NIMD_CENGAL_VERSION_MAJOR* = 4
const NIMD_CENGAL_VERSION_MINOR* = 4
const NIMD_CENGAL_VERSION_MICRO* = 0
const apiEndpointDoc* = """URL: https://api.production.com; Path: \usr\bin"""
