"""
Package greeter provides a simple greeting function

"""
# python wrapper for package cengal/_examples/ex_golang/versions/v_0/compilable/src within overall package greeter
# This is what you import to use the package.
# File is generated by gopy. Do not edit.
# gopy build -output=compiled -vm=/usr/bin/python3 ./src ./compile_time_py_definitions

# the following is required to enable dlopen to open the _go.so file
import os,sys,inspect,collections
try:
	import collections.abc as _collections_abc
except ImportError:
	_collections_abc = collections

cwd = os.getcwd()
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
os.chdir(currentdir)
from . import _greeter
from . import go

os.chdir(cwd)

# to use this code in your end-user python file, import it as follows:
# from greeter import greeter
# and then refer to everything using greeter. prefix
# packages imported by this package listed below:

from . import compile_time_py_definitions



# ---- Types ---


#---- Enums from Go (collections of consts with same type) ---


#---- Constants from Go: Python can only ask that you please don't change these! ---


# ---- Global Variables: can only use functions to access ---


# ---- Interfaces ---


# ---- Structs ---


# ---- Slices ---


# ---- Maps ---


# ---- Constructors ---


# ---- Functions ---
def Greet(name, goRun=False):
	"""Greet(str name) 
	
	Greet takes a name and prints a greeting message
	"""
	_greeter.greeter_Greet(name, goRun)
def Greet2(name, goRun=False):
	"""Greet2(str name) 
	
	Greet takes a name and prints a greeting message
	"""
	_greeter.greeter_Greet2(name, goRun)


